
#include "ggiterm.h"
#include <stdio.h>


#ifndef HAVE_FREETYPE

int
FreetypeInit (char *font_file, int fontsize, int (**ggitermWriteChar) (unsigned long))
{
	fprintf (stdout, "FreetypeInit: Freetype support not compiled in.\n");
	return 1;
}


void
FreetypeExit ()
{
}


#else /* HAVE_FREETYPE */

/* GGI */
#include <ggi/ggi.h>

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

/* Global variables */
FT_Library library;
FT_Face face;

/* External variables */
extern ggi_visual_t vis;
extern ggi_pixel fgcolor, bgcolor;
extern int cursor_x, cursor_y, cursor_x0, cursor_y0, cell_w, cell_h;
extern int blink_mode, bold_mode, invisible_mode, reverse_mode, underline_mode, altcharset_mode;

/* External functions */
void ggitermNextPos ();
unsigned long AsciiGetAltChar (unsigned long);


int
FreetypeValidateFont ()
{
	debug (DEBUG_INFO, "FreetypeValidateFont: Font family: %s\n", face->family_name);
	if (FT_IS_SCALABLE (face))
	{
		debug (DEBUG_INFO, "FreetypeValidateFont: font is scalable\n");
	}
	else
	{
		debug (DEBUG_INFO,    "FreetypeValidateFont: font is not scalable\n");
	}
	if (FT_IS_FIXED_WIDTH (face))
	{
		debug (DEBUG_INFO,
		       "FreetypeValidateFont: font is fixed width\n");
	}
	else
	{
		debug (DEBUG_INFO,
		       "FreetypeValidateFont: font is not fixed width\n");
		return 1;
	}
	debug (DEBUG_INFO, "FreetypeValidateFont: %ld glyphs in font\n",
	       face->num_glyphs);
	debug (DEBUG_INFO,
	       "FreetypeValidateFont: global metrics in font units:\n");
	debug (DEBUG_INFO,
	       "FreetypeValidateFont: BBox width  = %ld to %ld (%ld)\n",
	       face->bbox.xMin, face->bbox.xMax,
	       face->bbox.xMax - face->bbox.xMin);
	debug (DEBUG_INFO,
	       "FreetypeValidateFont: BBox height = %ld to %ld (%ld)\n",
	       face->bbox.yMin, face->bbox.yMax,
	       face->bbox.yMax - face->bbox.yMin);
	debug (DEBUG_INFO, "FreetypeValidateFont: text height = %d\n",
	       face->height);
	debug (DEBUG_INFO,
	       "FreetypeValidateFont: maximum advance width = %d\n",
	       face->max_advance_width);
	debug (DEBUG_INFO, "FreetypeValidateFont: units per EM = %d\n",
	       face->units_per_EM);
	debug (DEBUG_INFO, "FreetypeValidateFont: ascender = %d\n",
	       face->ascender);
	return 0;
}


unsigned long FreetypeGetAltChar (unsigned long charcode)
{
	unsigned long saved_charcode;
	
	saved_charcode = charcode;
	switch (charcode)
	{
		case 0x10: charcode = 0x2192; break;
		case 0x11: charcode = 0x2190; break;
		case 0x18: charcode = 0x2191; break;
		case 0x19: charcode = 0x2193; break;
		case 0xDB: charcode = 0x2588; break;
		case 0x04:
			charcode = 0x25C6;
			if (!FT_Get_Char_Index (face, charcode)) charcode = 0x266;
			break;
		case 0xB1: charcode = 0x2592; break;
		case 0xF8: charcode = 0x00B0; break;
		case 0xF1:
			charcode = 0x2213;
			if (!FT_Get_Char_Index (face, charcode)) charcode = 0x0B1;
			break;
		case 0xB0: charcode = 0x25A3; break;
		case 0xD9: charcode = 0x2518; break;
		case 0xBF: charcode = 0x2510; break;
		case 0xDA: charcode = 0x250C; break;
		case 0xC0: charcode = 0x2514; break;
		case 0xC5: charcode = 0x253C; break;
		case 0x7E: charcode = 0x23BA; break;
		case 0xC4: charcode = 0x2500; break;
		case 0x5F: charcode = 0x23BD; break;
		case 0xC3: charcode = 0x251C; break;
		case 0xB4: charcode = 0x2524; break;
		case 0xC1: charcode = 0x2534; break;
		case 0xC2: charcode = 0x252C; break;
		case 0xB3: charcode = 0x2502; break;
		case 0xF3: charcode = 0x2264; break;
		case 0xF2: charcode = 0x2265; break;
		case 0xE3: charcode = 0x03C0; break;
		case 0xD8: charcode = 0x2260; break;
		case 0x9C: charcode = 0x00A3; break;
		case 0xFE: charcode = 0x2022; break;
	}
	if (!FT_Get_Char_Index (face, charcode)) charcode = AsciiGetAltChar (saved_charcode);
	return charcode;
}

int
FreetypeWriteChar (unsigned long charcode)
{
	int x, y, colorlen, err, mypixelsize;
	unsigned char *tab;
	ggi_color *colortab;
	ggi_color bg, fg;
	ggi_pixel cur_bgcolor, cur_fgcolor;
	const ggi_pixelformat *mypixelformat;
	void *mypixelbuf;

	if (altcharset_mode)
	{
		charcode = FreetypeGetAltChar (charcode);
	}
	if (reverse_mode)
	{
		cur_bgcolor = fgcolor;
		cur_fgcolor = bgcolor;
	}
	else
	{
		cur_bgcolor = bgcolor;
		cur_fgcolor = fgcolor;
	}
	
	err = FT_Load_Char (face, charcode, FT_LOAD_RENDER);
	/*err = FT_Load_Char (face, charcode, FT_LOAD_MONOCHROME); */
	if (err)
	{
		fprintf (stderr,
			 "FreetypeWriteChar: No charmap defined or unknown character code\n");
		return 1;
	}
	debug (DEBUG_DETAIL, "FreetypeWriteChar: Glyph rendered to bitmap\n");

	debug (DEBUG_TOTAL, "FreetypeWriteChar: Bitmap fields:\n");
	debug (DEBUG_TOTAL, "FreetypeWriteChar: %d rows %d pixels wide\n",
	       face->glyph->bitmap.rows, face->glyph->bitmap.width);
	debug (DEBUG_TOTAL, "FreetypeWriteChar: pitch = %d\n",
	       face->glyph->bitmap.pitch);
	debug (DEBUG_TOTAL, "FreetypeWriteChar: Number of grays: %d\n",
	       face->glyph->bitmap.num_grays);

	debug (DEBUG_TOTAL, "FreetypeWriteChar: Bitmap values:\n");
	tab = face->glyph->bitmap.buffer;
	colortab =
		malloc (face->glyph->bitmap.rows * face->glyph->bitmap.width *
			sizeof (ggi_color));
	colorlen = 0;
	mypixelformat = ggiGetPixelFormat (vis);
	mypixelsize = mypixelformat->size;
	mypixelbuf =
		malloc (mypixelsize * face->glyph->bitmap.rows *
			face->glyph->bitmap.width);
	ggiUnmapPixel (vis, cur_bgcolor, &bg);
	ggiUnmapPixel (vis, cur_fgcolor, &fg);
	for (y = 0; y < face->glyph->bitmap.rows; y++)
	{
		debug (DEBUG_TOTAL, "FreetypeWriteChar: %3d: ", y);
		for (x = 0; x < face->glyph->bitmap.width; x++)
		{
			debug (DEBUG_TOTAL, " %3d", tab[x]);
			colortab[colorlen].r = bg.r + ((fg.r-bg.r) * tab[x]) / 256;
			colortab[colorlen].g = bg.g + ((fg.g-bg.g) * tab[x]) / 256;
			colortab[colorlen].b = bg.b + ((fg.b-bg.b) * tab[x]) / 256;
			colorlen++;
		}
		debug (DEBUG_TOTAL, "\n");
		tab += face->glyph->bitmap.pitch;
	}

	err = ggiPackColors (vis, mypixelbuf, colortab, colorlen);
	free (colortab);
	if (err)
	{
		fprintf (stderr, "FreetypeWriteChar: PackColors failed\n");
		free (mypixelbuf);
		return 1;
	}
	debug (DEBUG_DETAIL, "FreetypeWriteChar: PackColors succeeded\n");

	/* Erase background */
	ggiSetGCForeground (vis, cur_bgcolor);
	ggiDrawBox (vis, cell_w*(int)(cursor_x/cell_w), cell_h*(int)(cursor_y/cell_h), cell_w, cell_h);
	ggiSetGCForeground (vis, cur_fgcolor);
	
	err = ggiPutBox (vis, cursor_x + face->glyph->bitmap_left,
			 cursor_y - face->glyph->bitmap_top,
			 face->glyph->bitmap.width, face->glyph->bitmap.rows,
			 mypixelbuf);
	free (mypixelbuf);
	if (err)
	{
		fprintf (stderr, "FreetypeWriteChar: PutBox failed\n");
		return 1;
	}
	debug (DEBUG_DETAIL, "FreetypeWriteChar: PutBox succeeded\n");
	return 0;
}

int
FreetypeInit (char *font_file, int fontsize, int (**ggitermWriteChar) (unsigned long))
{
	int err;

	err = FT_Init_FreeType (&library);
	if (err)
	{
		fprintf (stderr,
			 "FreetypeInit: an error occurred during library initialization ...\n");
		return 1;
	}
	debug (DEBUG_INFO, "FreetypeInit: library initialized\n");

	err = FT_New_Face (library, font_file, 0, &face);
	if (err == FT_Err_Unknown_File_Format)
	{
		fprintf (stderr,
			 "FreetypeInit: the font file (%s) could be opened and read, but it appears that its font format is unsupported ...\n",
			 font_file);
		FT_Done_FreeType (library);
		return 1;
	}
	else if (err)
	{
		fprintf (stderr,
			 "FreetypeInit: the font file (%s) could not be opened or read, or simply is broken\n",
			 font_file);
		FT_Done_FreeType (library);
		return 1;

	}
	debug (DEBUG_INFO,
	       "FreetypeInit: font file (%s) successfully opened\n",
	       font_file);
	err = FreetypeValidateFont ();
	if (err)
	{
		fprintf (stderr, "FreetypeInit: Font not suitable!\n");
		FT_Done_FreeType (library);
		return 1;
	}
	err = FT_Set_Pixel_Sizes (face, fontsize, 0);
	if (err)
	{
		fprintf (stderr, "FreetypeInit: Couldn't set pixel size\n");
		FT_Done_FreeType (library);
		return 1;
	}
	cell_w = fontsize * (face->bbox.xMax -
			     face->bbox.xMin) / face->units_per_EM;
	cell_h = fontsize * face->height / face->units_per_EM;
	debug (DEBUG_INFO,
	       "FreetypeInit: Cell size successfully set to %dx%d\n", cell_w,
	       cell_h);
	cursor_x0 = 0;
	cursor_y0 = face->ascender * fontsize / face->units_per_EM;
	cursor_x = cursor_x0;
	cursor_y = cursor_y0;
	*ggitermWriteChar = &FreetypeWriteChar;
	debug (DEBUG_INFO,
	       "FreetypeInit: New handler ggitermWriteChar   -> FreetypeWriteChar\n");
	return 0;
}

void
FreetypeExit ()
{
	FT_Done_FreeType (library);
}

#endif /* HAVE_FREETYPE */


#ifdef DEBUG
void
FreetypeShowCharmap ()
{
#ifdef HAVE_FREETYPE
	FT_ULong charcode;
	FT_UInt gindex;

	if (debuglevel < DEBUG_INFO) return;
	
	charcode = FT_Get_First_Char (face, &gindex);
	while (gindex != 0)
	{
		FreetypeWriteChar (charcode);
		ggitermNextPos ();
		charcode = FT_Get_Next_Char (face, charcode, &gindex);
	}
#endif /* HAVE_FREETYPE */
}
#endif /* DEBUG */
