

#include "ggiterm.h"

#include <string.h>

/* For gettimeofday */
#include <sys/time.h>
#define BLINK_TIME 500

/* GGI */
#include <ggi/ggi.h>
#define GGI_EVMASK	(emKeyPress|emKeyRepeat)


/* Global variables */
ggi_visual_t vis;
int vis_w, vis_h;
int cursor_x, cursor_y, cursor_x0, cursor_y0, cell_w, cell_h;
ggi_pixel black, white, green;
ggi_pixel bgcolor, fgcolor;
int blink_mode = 0, bold_mode = 0, invisible_mode = 0, reverse_mode = 0, underline_mode = 0, altcharset_mode = 0;




unsigned long AsciiGetAltChar (unsigned long charcode)
{
		switch (charcode)
		{
			case 0x10: charcode = 0x3E; break;
			case 0x11: charcode = 0x3C; break;
			case 0x18: charcode = 0x5E; break;
			case 0x19: charcode = 0x76; break;
			case 0xDB: charcode = 0x23; break;
			case 0x04: charcode = 0x2B; break;
			case 0xB1: charcode = 0x3A; break;
			case 0xF8: charcode = 0x5C; break;
			case 0xF1: charcode = 0x23; break;
			case 0xB0: charcode = 0x23; break;
			case 0xD9: charcode = 0x2B; break;
			case 0xBF: charcode = 0x2B; break;
			case 0xDA: charcode = 0x2B; break;
			case 0xC0: charcode = 0x2B; break;
			case 0xC5: charcode = 0x2B; break;
			case 0x7E: charcode = 0x7E; break;
			case 0xC4: charcode = 0x2D; break;
			case 0x5F: charcode = 0x5F; break;
			case 0xC3: charcode = 0x2B; break;
			case 0xB4: charcode = 0x2B; break;
			case 0xC1: charcode = 0x2B; break;
			case 0xC2: charcode = 0x2B; break;
			case 0xB3: charcode = 0x7C; break;
			case 0xF3: charcode = 0x3C; break;
			case 0xF2: charcode = 0x3E; break;
			case 0xE3: charcode = 0x2A; break;
			case 0xD8: charcode = 0x21; break;
			case 0x9C: charcode = 0x66; break;
			case 0xFE: charcode = 0x6F; break;
		}
		return charcode;
}


void
ggitermBlinkCursor (int force)
{
	static struct timeval last_blink;
	static int state = 0;
	static int saved_x, saved_y;
	static char *buf;
	struct timeval current_time;
	long millisec;
	int mypixelsize;
	const ggi_pixelformat *mypixelformat;
		
	gettimeofday (&current_time, NULL);
	millisec = (current_time.tv_sec - last_blink.tv_sec) * 1000 + (current_time.tv_usec - last_blink.tv_usec)/1000;
	if (force==1 && state==1)
	{
		millisec = BLINK_TIME + 1;
	}
	if (force==2 && state==0)
	{
		millisec = BLINK_TIME+1;
	}
	if (millisec > BLINK_TIME)
	{
		if (state == 0)
		{
			mypixelformat = ggiGetPixelFormat (vis);
			mypixelsize = mypixelformat->size;
			buf = malloc (cell_w * cell_h * mypixelsize);
			ggiGetBox (vis, cell_w*(int)(cursor_x/cell_w), cell_h*(int)(cursor_y/cell_h), cell_w, cell_h, buf);
			ggiDrawBox (vis, cell_w*(int)(cursor_x/cell_w), cell_h*(int)(cursor_y/cell_h), cell_w, cell_h);
			saved_x = cursor_x;
			saved_y = cursor_y;
			state = 1;
		}
		else
		{
			ggiPutBox (vis, cell_w*(int)(saved_x/cell_w), cell_h*(int)(saved_y/cell_h), cell_w, cell_h, buf);
			free (buf);
			state = 0;
		}
		last_blink = current_time;
	}
}

void
ggitermDeleteChars (unsigned int nb_chars)
{
	int x, y;
	
	x = cell_w * (int)(cursor_x / cell_w);
	y = cell_h * (int)(cursor_y / cell_h);
	/* Copy the characters between current+nb_chars and EOL to current position */
	ggiCopyBox (vis, x+(nb_chars*cell_w), y, vis_w-x-(nb_chars*cell_w), cell_h, x, y);
	/* Is it necessary to blank the nb_chars last cells? */
}

void
ggitermInsertChars (unsigned int nb_chars)
{
	int x, y;
	
	x = cell_w * (int)(cursor_x / cell_w);
	y = cell_h * (int)(cursor_y / cell_h);
	/* First, copy the rest of the line nb_chars to the right */
	ggiCopyBox (vis, x, y, vis_w-x, cell_h, x+(nb_chars*cell_w), y);
	/* Then blank the newly created char cells */
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, x, y, nb_chars*cell_w, cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermInsertLines (unsigned int nb_lines)
{
	int y;
	
	y = cell_h * (int)(cursor_y / cell_h);
	/* First, copy the rest of the screen nb_lines lower */
	ggiCopyBox (vis, 0, y, vis_w, vis_h-y, 0, y+nb_lines*cell_h);
	/* Then blank the newly created lines */
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, y, vis_w, cell_h*nb_lines);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermDeleteLines (unsigned int nb_lines)
{
	int y;
	
	y = cell_h * (int)(cursor_y / cell_h);
	/* First, copy the section starting nb_lines lower to the current position */
	ggiCopyBox (vis, 0, y+nb_lines*cell_h, vis_w, vis_h-y-nb_lines*cell_h, 0, y);
	/* Then blank the lines created at the bottom of the screen */
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, vis_h-nb_lines*cell_h, vis_w, nb_lines*cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermEraseNChars (unsigned int nb_chars)
{
	int x, y;
	
	x = cell_w * (int)(cursor_x / cell_w);
	y = cell_h * (int)(cursor_y / cell_h);
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, x, y, nb_chars*cell_w, cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermEraseToEOL ()
{
	int x, y;
	
	x = cell_w * (int)(cursor_x / cell_w);
	y = cell_h * (int)(cursor_y / cell_h);
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, x, y, vis_w-x, cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermEraseFromBOL ()
{
	int x, y;
	
	x = cell_w * (int)(cursor_x / cell_w + 1);
	y = cell_h * (int)(cursor_y / cell_h);
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, y, x, cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermEraseToEOD ()
{
	int y;
	
	ggitermEraseToEOL ();
	y = cell_h * (int)(cursor_y / cell_h + 1);
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, y, vis_w, vis_h-y);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermScrollUp (unsigned int nb_lines)
{
	ggiCopyBox (vis, 0, 0, vis_w, vis_h, 0, -(nb_lines*cell_h));
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, vis_h - nb_lines*cell_h, vis_w, nb_lines*cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermScrollDown (unsigned int nb_lines)
{
	ggiCopyBox (vis, 0, 0, vis_w, vis_h, 0, nb_lines*cell_h);
	ggiSetGCForeground (vis, bgcolor);
	ggiDrawBox (vis, 0, 0, vis_w, nb_lines*cell_h);
	ggiSetGCForeground (vis, fgcolor);
}

void
ggitermNextPos ()
{
	cursor_x += cell_w;
	if (cursor_x >= (int)(vis_w / cell_w) * cell_w)
	{
		cursor_x = cursor_x0;
		cursor_y += cell_h;
		if (cursor_y >= (int)(vis_h / cell_h) * cell_h)
		{
			ggitermScrollUp (1);
			cursor_y = ((int)(vis_h / cell_h) - 1) * cell_h + cursor_y0;
		}
	}
}


int
GGIWriteChar (unsigned long charcode)
{
	ggi_pixel cur_bgcolor, cur_fgcolor;
	
	if (altcharset_mode)
	{
		charcode = AsciiGetAltChar (charcode);
	}
	if (reverse_mode)
	{
		cur_bgcolor = fgcolor;
		cur_fgcolor = bgcolor;
	}
	else
	{
		cur_bgcolor = bgcolor;
		cur_fgcolor = fgcolor;
	}
	ggiSetGCForeground (vis, cur_fgcolor);
	ggiSetGCBackground (vis, cur_bgcolor);
	ggiPutc (vis, cursor_x, cursor_y, (char) charcode);
	return 0;
}



void
HandleGGIEvents (char *buffer, int *index)
{
	struct timeval tv;
	ggi_event event;
	int i;

	tv.tv_sec = 0;
	tv.tv_usec = 0;

	if (ggiEventPoll (vis, GGI_EVMASK, &tv))
	{
		debug (DEBUG_DETAIL,
		       "HandleGGIEvents: Enter: inputbuffer index: %d\n",
		       *index);
		i = ggiEventsQueued (vis, GGI_EVMASK);
		debug (DEBUG_DETAIL,
		       "HandleGGIEvents: Number of GGI events queued: %d\n",
		       i);
		/* *index must be <=7 since escape sequences can be up to 3 chars long */
		while (i && *index <= 7)
		{
			ggiEventRead (vis, &event, GGI_EVMASK);
			i--;
			switch (event.any.type)
			{
			case evKeyRepeat:
			case evKeyPress:
				debug (DEBUG_INFO, "GGI event fields:\n");
				debug (DEBUG_INFO, "\t- Key number (scancode): %d\n", event.key.button);
				debug (DEBUG_INFO, "\t- Key label:             %c (%d)\n", event.key.label, event.key.label);
				debug (DEBUG_INFO, "\t- Symbol produced:       %c (%d)\n", event.key.sym, event.key.sym);
				debug (DEBUG_INFO, "\t- Modifiers:             ");
				if (event.key.modifiers & GII_MOD_SHIFT) debug (DEBUG_INFO, "SHIFT ");
				if (event.key.modifiers & GII_MOD_CTRL) debug (DEBUG_INFO, "CTRL ");
				if (event.key.modifiers & GII_MOD_ALT) debug (DEBUG_INFO, "ALT ");
				if (event.key.modifiers & GII_MOD_META) debug (DEBUG_INFO, "META ");
				if (event.key.modifiers & GII_MOD_SUPER) debug (DEBUG_INFO, "SUPER ");
				if (event.key.modifiers & GII_MOD_HYPER) debug (DEBUG_INFO, "HYPER ");
				if (event.key.modifiers & GII_MOD_ALTGR) debug (DEBUG_INFO, "ALTGR ");
				if (event.key.modifiers & GII_MOD_CAPS) debug (DEBUG_INFO, "CAPS ");
				if (event.key.modifiers & GII_MOD_NUM) debug (DEBUG_INFO, "NUM ");
				if (event.key.modifiers & GII_MOD_SCROLL) debug (DEBUG_INFO, "SCROLL ");
				debug (DEBUG_INFO, "\n");
				
				switch (event.key.sym)
				{
					case GIIK_Enter:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [ENTER]\n");
						buffer[*index] = '\n';
						(*index)++;
						break;
					case GIIK_Left:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [LEFT]\n");
						memcpy (buffer+(*index), "\033[D", 3);
						(*index) += 3;
						break;
					case GIIK_Right:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [RIGHT]\n");
						memcpy (buffer+(*index), "\033[C", 3);
						(*index) += 3;
						break;
					case GIIK_Up:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [UP]\n");
						memcpy (buffer+(*index), "\033[A", 3);
						(*index) += 3;
						break;
					case GIIK_Down:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [DOWN]\n");
						memcpy (buffer+(*index), "\033[B", 3);
						(*index) += 3;
						break;
					case GIIK_Home:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [HOME]\n");
						memcpy (buffer+(*index), "\033[H", 3);
						(*index) += 3;
						break;
					case GIIK_Insert:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [INSERT]\n");
						memcpy (buffer+(*index), "\033[L", 3);
						(*index) += 3;
						break;
					/* The following are not strictly needed: the cases are handled
					by the default section but this way debugging output is not messed up! */
					case GIIUC_BackSpace:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [BACKSPACE]\n");
						buffer[*index] = 0x08;
						(*index)++;
						break;
					case GIIUC_Tab:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed [TAB]\n");
						buffer[*index] = 0x09;
						(*index)++;
						break;
					default:
						debug (DEBUG_INFO, "HandleGGIEvents: You pressed the key: '%c'\n",
				       		GII_KVAL (event.key.sym));
						buffer[*index] = GII_KVAL (event.key.sym);
						(*index)++;
				}
				break;
			}
		}
		debug (DEBUG_DETAIL,
		       "HandleGGIEvents: Exit: inputbuffer index: %d\n",
		       *index);
	}
}


int
GGIInit_graphicsmode (void)
{
	int err;
	ggi_mode sug_mode;
	ggi_color color;

	if (ggiInit () < 0)
	{
		fprintf (stderr, "GGIInit_graphicsmode: Cannot initalize LibGGI!\n");
		return 1;
	}
	debug (DEBUG_INFO, "GGIInit_graphicsmode: LibGGI initialization successful.\n");

	vis = ggiOpen (NULL);
	if (vis == NULL)
	{
		fprintf (stderr, "GGIInit_graphicsmode: Couldn't open default visual!\n");
		ggiExit ();
		return 1;
	}
	debug (DEBUG_INFO, "GGIInit_graphicsmode: Default visual opened\n");
	
	err = ggiCheckGraphMode (vis, GGI_AUTO, GGI_AUTO, GGI_AUTO, GGI_AUTO,
				 GT_AUTO, &sug_mode);
	if (err)
	{
		fprintf (stderr,
			 "GGIInit_graphicsmode: No graphics mode available!\n");
		return 1;
	}
	debug (DEBUG_INFO,
	       "GGIInit_graphicsmode: Automatic graphics mode check succeeded. Suggested mode is: ");
#ifdef DEBUG
	if (DEBUG_INFO <= debuglevel)
		ggiFPrintMode (stdout, &sug_mode);
#endif /* DEBUG */
	debug (DEBUG_INFO, "\n");
	debug (DEBUG_INFO,
	       "GGIInit_graphicsmode: Physical resolution: %dx%d mm\n",
	       sug_mode.size.x, sug_mode.size.y);

	debug (DEBUG_INFO, "GGIInit_graphicsmode: Setting graphics mode\n");
	err = ggiSetMode (vis, &sug_mode);
	if (err)
	{
		fprintf (stderr,
			 "GGIInit_graphicsmode: Set mode failed although check was OK!\n");
		return 1;
	}
	vis_w = sug_mode.visible.x;	/* in pixels */
	vis_h = sug_mode.visible.y;	/* in pixels */
	ggiGetCharSize (vis, &cell_w, &cell_h);


	color.r = 0xFFFF;
	color.g = 0xFFFF;
	color.b = 0xFFFF;
	white = ggiMapColor (vis, &color);
	fgcolor = white;
	color.r = 0x0000;
	color.g = 0x0000;
	color.b = 0x0000;
	black = ggiMapColor (vis, &color);
	bgcolor = black;
	color.r = 0x0000;
	color.g = 0xFFFF;
	color.b = 0x0000;
	green = ggiMapColor (vis, &color);
	debug (DEBUG_INFO,
	       "GGIInit_graphicsmode: Colors: black=%d, white=%d, green=%d\n",
	       black, white, green);

	ggiSetGCBackground (vis, bgcolor);
	ggiSetGCForeground (vis, fgcolor);
	return 0;
}

int
GGIInit_textmode (void)
{
	int err;
	ggi_mode sug_mode;
	ggi_color color;
	ggi_pixel green;

	err = ggiCheckTextMode (vis, GGI_AUTO, GGI_AUTO, GGI_AUTO, GGI_AUTO,
				GGI_AUTO, GGI_AUTO, GT_AUTO, &sug_mode);
	if (err)
	{
		fprintf (stderr,
			 "GGIInit_textmode: Automatic text mode check failed!\n");
		return 1;
	}
	debug (DEBUG_INFO,
	       "GGIInit_textmode: Automatic text mode check succeeded. Suggested mode is: ");
#ifdef DEBUG
	if (DEBUG_INFO <= debuglevel)
		ggiFPrintMode (stdout, &sug_mode);
#endif /* DEBUG */
	debug (DEBUG_INFO, "\n");

	debug (DEBUG_INFO, "GGIInit_textmode: Setting text mode\n");
	err = ggiSetMode (vis, &sug_mode);
	if (err)
	{
		fprintf (stderr,
			 "GGIInit_textmode: Set mode failed although check was OK!\n");
		return 1;
	}
	vis_w = sug_mode.visible.x;
	vis_h = sug_mode.visible.y;
	ggiGetCharSize (vis, &cell_w, &cell_h);

	color.r = 0xFFFF;
	color.g = 0xFFFF;
	color.b = 0xFFFF;
	white = ggiMapColor (vis, &color);
	fgcolor = white;
	color.r = 0x0000;
	color.g = 0x0000;
	color.b = 0x0000;
	black = ggiMapColor (vis, &color);
	bgcolor = black;
	color.r = 0x0000;
	color.g = 0xFFFF;
	color.b = 0x0000;
	green = ggiMapColor (vis, &color);
	debug (DEBUG_INFO,
	       "GGIInit_textmode: Colors: black=%d, white=%d, green=%d\n",
	       black, white, green);
	ggiSetGCBackground (vis, bgcolor);
	ggiSetGCForeground (vis, fgcolor);
	return 0;
}

#ifdef DEBUG
void ShowGrid (void)
{
	int x, y;
	
	if (debuglevel < DEBUG_INFO) return;
		
	ggiSetGCForeground (vis, green);
	for (x = 0; x < vis_w; x += cell_w)
	{
		ggiDrawVLine (vis, x, 0, vis_h);
	}
	for (y = 0; y < vis_h; y += cell_h)
	{
		ggiDrawHLine (vis, 0, y, vis_w);
	}
}
#endif /* DEBUG */

void GGIExit (void)
{
	ggiClose (vis);
	ggiExit ();
}
