
#include "ggiterm.h"
#include <iconv.h>
#include <errno.h>
#include <stdio.h>


/* Don't forget to specify the endianness
 * ie UTF32LE and NOT UTF32
 * if not, iconv will produce a BOM code at the
 * beginning of the output stream, which will mess up
 * the conversion algorithm!
 *
 * This forces us to determine
 * endianness in the configure process
 */
#ifdef WORDS_BIGENDIAN
 /*#define ICONV_TO "UTF32BE" */
#define ICONV_TO "UNICODEBIG"
#else
 /*#define ICONV_TO "UTF32LE" */
#define ICONV_TO "UNICODELITTLE"
#endif /* WORDS_BIGENDIAN */


#ifndef HAVE_ICONV


int
IconvInit (int (**ggitermGetNextChar) (unsigned short*, char**, size_t*))
{
	fprintf (stdout, "IconvInit: Iconv support not compiled in.\n");
	return 1;
}

void IconvExit ()
{
}

#else /* HAVE_ICONV */


iconv_t iconv_handler;
char* encoding_from = "latin1";


int
IconvGetNextChar (unsigned short *charcode, char **inputbuf,
		  size_t * inputsize)
{
	char *outputbuf;
	size_t outputsize;

	outputbuf = (char *) charcode;
	outputsize = sizeof (unsigned short);
	debug (DEBUG_DETAIL,
	       "IconvGetNextChar: Remaining before conversion: input=%d output=%d\n",
	       *inputsize, outputsize);
	/*debug (DEBUG_DETAIL, "IconvGetNextChar: '%c'\tencoding=%d\n",
	       (*inputbuf)[0], (unsigned char) (*inputbuf)[0]);*/
	/* It seems we can't rely on iconv() returning -1 when an error occured
	   so we set errno to 0 and see whether it was indeed changed */
	errno = 0;
	iconv (iconv_handler, (ICONV_CONST char **) inputbuf, inputsize,
		     &outputbuf, &outputsize);
	debug (DEBUG_DETAIL,
	       "IconvGetNextChar: Remaining after conversion: input=%d output=%d\n",
	       *inputsize, outputsize);
	if (errno)
	{
		switch (errno)
		{
			case E2BIG:
				return 0;
			case EILSEQ:
				/* invalid character encountered */
				debug (DEBUG_INFO,
				       "IconvGetNextChar: Invalid byte, skipping it\n");
				(*inputbuf)++;
				(*inputsize)--;
				return 1;
			case EINVAL:
				/* incomplete input */
				debug (DEBUG_INFO,
				       "IconvGetNextChar: Incomplete input\n");
				return 2;
			default:
				perror ("ggiterm: iconv");
				return 1;
		}
	}
	return 0;
}


int
IconvInit (int (**ggitermGetNextChar) (unsigned short*, char**, size_t*))
{
	iconv_handler = iconv_open (ICONV_TO, encoding_from);
	if (iconv_handler == (iconv_t) (-1))
	{
		perror ("ggiterm: iconv_open");
		return 1;
	}
	*ggitermGetNextChar = &IconvGetNextChar;
	debug (DEBUG_INFO, "IconvInit: Converting from %s to %s\n",
	       encoding_from, ICONV_TO);
	debug (DEBUG_INFO, "IconvInit: New handler ggitermGetNextChar -> IconvGetNextChar\n");
	return 0;
}


void IconvExit ()
{
	iconv_close (iconv_handler);
}

#endif /* HAVE_ICONV */
